@@ -228,6 +228,11 @@ struct MVMBoolificationSpec {
     MVMuint32  mode;
 };
 
+/* Constant for incrementing the type cache ID for new STables. This leaves
+ * the lowest bits free for caches to attach flags (of note, the multi
+ * dispatch cache). */
+#define MVM_TYPE_CACHE_ID_INCR 128
+
 /* S-table, representing a meta-object/representation pairing. Note that the
  * items are grouped in hope that it will pack decently and do decently in
  * terms of cache lines. */
