@@ -1,5 +1,5 @@
 #include "moar.h"
-#include "math.h"
+#include <math.h>
 
 /* This representation's function pointer table. */
 static const MVMREPROps this_repr;
@@ -114,6 +114,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
     /* If we have no attributes in the index mapping, then just the header. */
     if (repr_data->name_to_index_mapping[0].class_key == NULL) {
         repr_data->struct_size = 1; /* avoid 0-byte malloc */
+        repr_data->struct_align = ALIGNOF(void *);
     }
 
     /* Otherwise, we need to compute the allocation strategy.  */
@@ -122,7 +123,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
         MVMint32 cur_size    = 0;
         /* The structure itself will be the multiple of its biggest element in size.
          * So we keep track of that biggest element. */
-        MVMint32 multiple_of = 0;
+        MVMint32 multiple_of = 1;
 
         /* Get number of attributes and set up various counters. */
         MVMint32 num_attrs        = MVM_repr_elems(tc, flat_list);
@@ -137,6 +138,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
         repr_data->struct_offsets      = (MVMint32 *)   MVM_malloc(info_alloc * sizeof(MVMint32));
         repr_data->flattened_stables   = (MVMSTable **) MVM_calloc(info_alloc, sizeof(MVMObject *));
         repr_data->member_types        = (MVMObject **) MVM_calloc(info_alloc, sizeof(MVMObject *));
+        repr_data->struct_align        = 0;
 
         /* Go over the attributes and arrange their allocation. */
         for (i = 0; i < num_attrs; i++) {
@@ -198,10 +200,22 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     if (inlined) {
                         MVMCStructREPRData *cstruct_repr_data = (MVMCStructREPRData *)STABLE(type)->REPR_data;
                         bits                                  = cstruct_repr_data->struct_size * 8;
-                        align                                 = cstruct_repr_data->struct_size;
+                        align                                 = cstruct_repr_data->struct_align;
                         repr_data->attribute_locations[i]    |= MVM_CSTRUCT_ATTR_INLINED;
                     }
                 }
+                else if (type_id == MVM_REPR_ID_MVMCPPStruct) {
+                    /* It's a CPPStruct. */
+                    repr_data->num_child_objs++;
+                    repr_data->attribute_locations[i] = (cur_obj_attr++ << MVM_CSTRUCT_ATTR_SHIFT) | MVM_CSTRUCT_ATTR_CPPSTRUCT;
+                    repr_data->member_types[i] = type;
+                    if (inlined) {
+                        MVMCPPStructREPRData *cppstruct_repr_data = (MVMCPPStructREPRData *)STABLE(type)->REPR_data;
+                        bits                                      = cppstruct_repr_data->struct_size * 8;
+                        align                                     = cppstruct_repr_data->struct_align;
+                        repr_data->attribute_locations[i]        |= MVM_CSTRUCT_ATTR_INLINED;
+                    }
+                }
                 else if (type_id == MVM_REPR_ID_MVMCUnion) {
                     /* It's a CUnion. */
                     repr_data->num_child_objs++;
@@ -210,7 +224,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     if (inlined) {
                         MVMCUnionREPRData *cunion_repr_data = (MVMCUnionREPRData *)STABLE(type)->REPR_data;
                         bits                                = cunion_repr_data->struct_size * 8;
-                        align                               = cunion_repr_data->struct_size;
+                        align                               = cunion_repr_data->struct_align;
                         repr_data->attribute_locations[i]  |= MVM_CSTRUCT_ATTR_INLINED;
                     }
                 }
@@ -222,7 +236,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                 }
                 else {
                     MVM_exception_throw_adhoc(tc,
-                        "CStruct representation only handles int, num, CArray, CPointer and CStruct");
+                        "CStruct representation only handles int, num, CArray, CPointer, CStruct, CPPStruct and CUnion");
                 }
             }
             else {
@@ -243,6 +257,9 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                 cur_size += align - cur_size % align;
             }
 
+            if (align > repr_data->struct_align)
+                repr_data->struct_align = align;
+
             repr_data->struct_offsets[i] = cur_size;
             cur_size += bits / 8;
 
@@ -397,6 +414,13 @@ static void get_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
                             else
                                 obj = MVM_nativecall_make_cstruct(tc, typeobj, cobj);
                         }
+                        else if(type == MVM_CSTRUCT_ATTR_CPPSTRUCT) {
+                            if (repr_data->attribute_locations[slot] & MVM_CSTRUCT_ATTR_INLINED)
+                                obj = MVM_nativecall_make_cppstruct(tc, typeobj,
+                                    (char *)body->cstruct + repr_data->struct_offsets[slot]);
+                            else
+                                obj = MVM_nativecall_make_cppstruct(tc, typeobj, cobj);
+                        }
                         else if(type == MVM_CSTRUCT_ATTR_CUNION) {
                             if (repr_data->attribute_locations[slot] & MVM_CSTRUCT_ATTR_INLINED)
                                 obj = MVM_nativecall_make_cunion(tc, typeobj,
@@ -503,6 +527,12 @@ static void bind_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
                                 "Can only store CStruct attribute in CStruct slot in CStruct");
                         cobj = ((MVMCStruct *)value)->body.cstruct;
                     }
+                    else if (type == MVM_CSTRUCT_ATTR_CPPSTRUCT) {
+                        if (REPR(value)->ID != MVM_REPR_ID_MVMCPPStruct)
+                            MVM_exception_throw_adhoc(tc,
+                                "Can only store CPPStruct attribute in CPPStruct slot in CStruct");
+                        cobj = ((MVMCPPStruct *)value)->body.cppstruct;
+                    }
                     else if (type == MVM_CSTRUCT_ATTR_CUNION) {
                         if (REPR(value)->ID != MVM_REPR_ID_MVMCUnion)
                             MVM_exception_throw_adhoc(tc,
