@@ -1,5 +1,4 @@
 #include "moar.h"
-#include "math.h"
 
 /* This representation's function pointer table. */
 static const MVMREPROps this_repr;
@@ -114,6 +113,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
     /* If we have no attributes in the index mapping, then just the header. */
     if (repr_data->name_to_index_mapping[0].class_key == NULL) {
         repr_data->struct_size = 1; /* avoid 0-byte malloc */
+        repr_data->struct_align = ALIGNOF(void *);
     }
 
     /* Otherwise, we need to compute the allocation strategy.  */
@@ -136,6 +136,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
         repr_data->struct_offsets      = (MVMint32 *)   MVM_malloc(info_alloc * sizeof(MVMint32));
         repr_data->flattened_stables   = (MVMSTable **) MVM_calloc(info_alloc, sizeof(MVMObject *));
         repr_data->member_types        = (MVMObject **) MVM_calloc(info_alloc, sizeof(MVMObject *));
+        repr_data->struct_align        = 0;
 
         /* Go over the attributes and arrange their allocation. */
         for (i = 0; i < num_attrs; i++) {
@@ -197,9 +198,22 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     if (inlined) {
                         MVMCStructREPRData *cstruct_repr_data = (MVMCStructREPRData *)STABLE(type)->REPR_data;
                         bits                                  = cstruct_repr_data->struct_size * 8;
+                        align                                 = cstruct_repr_data->struct_align;
                         repr_data->attribute_locations[i]    |= MVM_CUNION_ATTR_INLINED;
                     }
                 }
+                else if (type_id == MVM_REPR_ID_MVMCPPStruct) {
+                    /* It's a CPPStruct. */
+                    repr_data->num_child_objs++;
+                    repr_data->attribute_locations[i] = (cur_obj_attr++ << MVM_CUNION_ATTR_SHIFT) | MVM_CUNION_ATTR_CPPSTRUCT;
+                    repr_data->member_types[i] = type;
+                    if (inlined) {
+                        MVMCPPStructREPRData *cppstruct_repr_data = (MVMCPPStructREPRData *)STABLE(type)->REPR_data;
+                        bits                                      = cppstruct_repr_data->struct_size * 8;
+                        align                                     = cppstruct_repr_data->struct_align;
+                        repr_data->attribute_locations[i]        |= MVM_CUNION_ATTR_INLINED;
+                    }
+                }
                 else if (type_id == MVM_REPR_ID_MVMCUnion) {
                     /* It's a CUnion. */
                     repr_data->num_child_objs++;
@@ -208,6 +222,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     if (inlined) {
                         MVMCUnionREPRData *cunion_repr_data = (MVMCUnionREPRData *)STABLE(type)->REPR_data;
                         bits                                = cunion_repr_data->struct_size * 8;
+                        align                               = cunion_repr_data->struct_align;
                         repr_data->attribute_locations[i]  |= MVM_CUNION_ATTR_INLINED;
                     }
                 }
@@ -219,7 +234,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                 }
                 else {
                     MVM_exception_throw_adhoc(tc,
-                        "CUnion representation only handles int, num, CArray, CPointer and CStruct");
+                        "CUnion representation only handles int, num, CArray, CPointer, CStruct, CPPStruct and CUnion");
                 }
             }
             else {
@@ -232,6 +247,9 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     "CUnion only supports native types that are a multiple of 8 bits wide (was passed: %"PRId32")", bits);
             }
 
+            if (align > repr_data->struct_align)
+                repr_data->struct_align = align;
+
             repr_data->struct_offsets[i] = 0;
             if (bits / 8 > total_size)
                 total_size = bits / 8;
@@ -410,6 +428,13 @@ static void get_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
                             else
                                 obj = MVM_nativecall_make_cstruct(tc, typeobj, cobj);
                         }
+                        else if(type == MVM_CUNION_ATTR_CPPSTRUCT) {
+                            if (repr_data->attribute_locations[slot] & MVM_CUNION_ATTR_INLINED)
+                                obj = MVM_nativecall_make_cppstruct(tc, typeobj,
+                                    (char *)body->cunion + repr_data->struct_offsets[slot]);
+                            else
+                                obj = MVM_nativecall_make_cppstruct(tc, typeobj, cobj);
+                        }
                         else if(type == MVM_CUNION_ATTR_CUNION) {
                             if (repr_data->attribute_locations[slot] & MVM_CUNION_ATTR_INLINED)
                                 obj = MVM_nativecall_make_cunion(tc, typeobj,
