@@ -33,7 +33,7 @@ MVM_STATIC_INLINE size_t indices_to_flat_index(MVMThreadContext *tc, MVMint64 nu
         else {
             MVM_exception_throw_adhoc(tc,
                 "Index %"PRId64" for dimension %"PRId64" out of range (must be 0..%"PRId64")",
-                index, i + 1, dim_size);
+                index, i + 1, dim_size - 1);
         }
     }
     return result;
@@ -703,6 +703,12 @@ static MVMuint64 elems(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, voi
 static MVMStorageSpec get_elem_storage_spec(MVMThreadContext *tc, MVMSTable *st) {
     MVMMultiDimArrayREPRData *repr_data = (MVMMultiDimArrayREPRData *)st->REPR_data;
     MVMStorageSpec spec;
+
+    /* initialise storage spec to default values */
+    spec.bits            = 0;
+    spec.align           = 0;
+    spec.is_unsigned     = 0;
+
     switch (repr_data->slot_type) {
         case MVM_ARRAY_STR:
             spec.inlineable      = MVM_STORAGE_SPEC_INLINED;
