@@ -5,36 +5,45 @@
  * bytecode. */
 static void prepare_and_verify_static_frame(MVMThreadContext *tc, MVMStaticFrame *static_frame) {
     MVMStaticFrameBody *static_frame_body = &static_frame->body;
+    MVMCompUnit        *cu                = static_frame_body->cu;
 
     /* Ensure the frame is fully deserialized. */
     if (!static_frame_body->fully_deserialized)
-        MVM_bytecode_finish_frame(tc, static_frame_body->cu, static_frame, 0);
+        MVM_bytecode_finish_frame(tc, cu, static_frame, 0);
 
-    /* Work size is number of locals/registers plus size of the maximum
-     * call site argument list. */
-    static_frame_body->work_size = sizeof(MVMRegister) *
-        (static_frame_body->num_locals + static_frame_body->cu->body.max_callsite_size);
+    /* Take compilation unit lock, to make sure we don't race to do the
+     * frame preparation/verification work. */
+    MVM_reentrantmutex_lock(tc, (MVMReentrantMutex *)cu->body.update_mutex);
+    if (static_frame->body.instrumentation_level == 0) {
+        /* Work size is number of locals/registers plus size of the maximum
+        * call site argument list. */
+        static_frame_body->work_size = sizeof(MVMRegister) *
+            (static_frame_body->num_locals + static_frame_body->cu->body.max_callsite_size);
 
-    /* Validate the bytecode. */
-    MVM_validate_static_frame(tc, static_frame);
+        /* Validate the bytecode. */
+        MVM_validate_static_frame(tc, static_frame);
 
-    /* Obtain an index to each threadcontext's lexotic pool table */
-    static_frame_body->pool_index = MVM_incr(&tc->instance->num_frames_run);
+        /* Obtain an index to each threadcontext's lexotic pool table */
+        static_frame_body->pool_index = MVM_incr(&tc->instance->num_frames_run);
 
-    /* Check if we have any state var lexicals. */
-    if (static_frame_body->static_env_flags) {
-        MVMuint8 *flags  = static_frame_body->static_env_flags;
-        MVMint64  numlex = static_frame_body->num_lexicals;
-        MVMint64  i;
-        for (i = 0; i < numlex; i++)
-            if (flags[i] == 2) {
-                static_frame_body->has_state_vars = 1;
-                break;
-            }
+        /* Check if we have any state var lexicals. */
+        if (static_frame_body->static_env_flags) {
+            MVMuint8 *flags  = static_frame_body->static_env_flags;
+            MVMint64  numlex = static_frame_body->num_lexicals;
+            MVMint64  i;
+            for (i = 0; i < numlex; i++)
+                if (flags[i] == 2) {
+                    static_frame_body->has_state_vars = 1;
+                    break;
+                }
+        }
+
+        /* Set its spesh threshold. */
+        static_frame_body->spesh_threshold = MVM_spesh_threshold(tc, static_frame);
     }
 
-    /* Set its spesh threshold. */
-    static_frame_body->spesh_threshold = MVM_spesh_threshold(tc, static_frame);
+    /* Unlock, now we're finished. */
+    MVM_reentrantmutex_unlock(tc, (MVMReentrantMutex *)cu->body.update_mutex);
 }
 
 /* When we don't match the current instrumentation level, we hit this. It may
@@ -58,36 +67,74 @@ static void instrumentation_level_barrier(MVMThreadContext *tc, MVMStaticFrame *
         MVM_profile_ensure_uninstrumented(tc, static_frame);
 }
 
-/* Increases the reference count of a frame. */
+/* Increases the reference count of a frame, and marks it as having been
+ * referenced by a garbage collectable object. Can only safely be used
+ * if you know the frame is referenced by at least one other frame or by
+ * being on the stack top. If there's a chance it may be referenced only
+ * by an object, use MVM_frame_acquire_ref to avoid races. */
 MVMFrame * MVM_frame_inc_ref(MVMThreadContext *tc, MVMFrame *frame) {
+    frame->refd_by_object = 1;
     MVM_incr(&frame->ref_count);
     return frame;
 }
 
+/* Increases the reference count of a frame. Can only safely be used when
+ * the thing that will refer to the frame is another frame, and when you
+ * know the frame is referenced by at least one other frame or by being on
+ * the stack top. If there's a chance it may be referenced only by an object,
+ * use MVM_frame_acquire_ref to avoid races. */
+MVMFrame * MVM_frame_inc_ref_by_frame(MVMThreadContext *tc, MVMFrame *frame) {
+    MVM_incr(&frame->ref_count);
+    return frame;
+}
+
+/* Takes a pointer to a location that points to an MVMFrame, and safely
+ * obtains a reference to the frame. If the frame should die while we
+ * try to do so, then it will dereference the frame pointer afresh and
+ * try again. */
+MVMFrame * MVM_frame_acquire_ref(MVMThreadContext *tc, MVMFrame **frame) {
+    while (1) {
+        MVMFrame *try = (MVMFrame *)MVM_load(frame);
+        if (try) {
+            AO_t ref_count = MVM_load(&(try->ref_count));
+            if (ref_count > 0)
+                if (MVM_trycas(&(try->ref_count), ref_count, ref_count + 1))
+                    return try;
+        }
+        else {
+            return NULL;
+        }
+    }
+}
+
 /* Decreases the reference count of a frame. If it hits zero, then we can
  * free it. Returns null for convenience. */
 MVMFrame * MVM_frame_dec_ref(MVMThreadContext *tc, MVMFrame *frame) {
     /* MVM_dec returns what the count was before it decremented it
      * to zero, so we look for 1 here. */
     while (MVM_decr(&frame->ref_count) == 1) {
-        MVMuint32 pool_index = frame->static_info->body.pool_index;
         MVMFrame *outer_to_decr = frame->outer;
 
         /* If there's a caller pointer, decrement that. */
         if (frame->caller)
             frame->caller = MVM_frame_dec_ref(tc, frame->caller);
 
-        /* Destroy the frame. */
-        if (frame->env) {
-            MVM_fixed_size_free(tc, tc->instance->fsa, frame->allocd_env,
-                frame->env);
-        }
+        /* Destroy the frame. If it was referenced by an object, we need to
+         * free it at the next safe point; the 0'd ref count must remain
+         * readable until then since we use it as a sentinel value to avoid
+         * anything acquiring a reference to a dead frame. */
         if (frame->work) {
             MVM_args_proc_cleanup(tc, &frame->params);
             MVM_fixed_size_free(tc, tc->instance->fsa, frame->allocd_work,
                 frame->work);
         }
-        MVM_fixed_size_free(tc, tc->instance->fsa, sizeof(MVMFrame), frame);
+        if (frame->env)
+            MVM_fixed_size_free_at_safepoint(tc, tc->instance->fsa,
+                frame->allocd_env, frame->env);
+        if (frame->refd_by_object)
+            MVM_fixed_size_free_at_safepoint(tc, tc->instance->fsa, sizeof(MVMFrame), frame);
+        else
+            MVM_fixed_size_free(tc, tc->instance->fsa, sizeof(MVMFrame), frame);
 
         /* Decrement any outer. */
         if (outer_to_decr)
@@ -157,8 +204,9 @@ static MVMFrame * create_context_only(MVMThreadContext *tc, MVMStaticFrame *stat
  * with a ref count of 1 due to being held by an SC. */
 MVMFrame * MVM_frame_create_context_only(MVMThreadContext *tc, MVMStaticFrame *static_frame,
         MVMObject *code_ref) {
-    return MVM_frame_inc_ref(tc,
-        create_context_only(tc, static_frame, code_ref, 0));
+    MVMFrame *f = create_context_only(tc, static_frame, code_ref, 0);
+    f->ref_count = 1;
+    return f;
 }
 
 /* Provides auto-close functionality, for the handful of cases where we have
@@ -183,11 +231,11 @@ static MVMFrame * autoclose(MVMThreadContext *tc, MVMStaticFrame *needed) {
         if (outer_code->body.outer &&
                 outer_code->body.outer->static_info->body.bytecode == needed->body.bytecode) {
             /* Yes, just take it. */
-            result->outer = MVM_frame_inc_ref(tc, outer_code->body.outer);
+            result->outer = MVM_frame_acquire_ref(tc, &(outer_code->body.outer));
         }
         else {
             /* Otherwise, recursively auto-close. */
-            result->outer = MVM_frame_inc_ref(tc, autoclose(tc, needed->body.outer));
+            result->outer = MVM_frame_inc_ref_by_frame(tc, autoclose(tc, needed->body.outer));
         }
     }
     return result;
@@ -197,12 +245,12 @@ static MVMFrame * autoclose(MVMThreadContext *tc, MVMStaticFrame *needed) {
 static MVMFrame * allocate_frame(MVMThreadContext *tc, MVMStaticFrameBody *static_frame_body,
                                  MVMSpeshCandidate *spesh_cand) {
     MVMFrame *frame = NULL;
-    MVMFrame *node;
     MVMint32  env_size, work_size;
 
     /* Allocate the frame. */
     frame = MVM_fixed_size_alloc(tc, tc->instance->fsa, sizeof(MVMFrame));
     frame->params.named_used = NULL;
+    frame->refd_by_object = 0;
 
     /* Ensure special return pointers, continuation tags, dynlex cache,
      * and return address are null. */
@@ -249,7 +297,8 @@ static MVMFrame * allocate_frame(MVMThreadContext *tc, MVMStaticFrameBody *stati
 void MVM_frame_invoke_code(MVMThreadContext *tc, MVMCode *code,
                            MVMCallsite *callsite, MVMint32 spesh_cand) {
     MVM_frame_invoke(tc, code->body.sf, callsite,  tc->cur_frame->args,
-                     code->body.outer, (MVMObject*)code, spesh_cand);
+        MVM_frame_acquire_ref(tc, &(code->body.outer)),
+        (MVMObject*)code, spesh_cand);
 }
 
 /* Takes a static frame and a thread context. Invokes the static frame. */
@@ -320,6 +369,30 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
                             match = 0;
                         break;
                     }
+                    case MVM_SPESH_GUARD_DC_CONC_RW: {
+                        if (STABLE(arg)->container_spec->can_store(tc, arg)) {
+                            MVMRegister dc;
+                            STABLE(arg)->container_spec->fetch(tc, arg, &dc);
+                            if (!dc.o || !IS_CONCRETE(dc.o) || STABLE(dc.o) != st)
+                                match = 0;
+                        }
+                        else {
+                            match = 0;
+                        }
+                        break;
+                    }
+                    case MVM_SPESH_GUARD_DC_TYPE_RW: {
+                        if (STABLE(arg)->container_spec->can_store(tc, arg)) {
+                            MVMRegister dc;
+                            STABLE(arg)->container_spec->fetch(tc, arg, &dc);
+                            if (!dc.o || IS_CONCRETE(dc.o) || STABLE(dc.o) != st)
+                                match = 0;
+                        }
+                        else {
+                            match = 0;
+                        }
+                        break;
+                    }
                     }
                     if (!match)
                         break;
@@ -394,9 +467,10 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
 
     /* Outer. */
     if (outer) {
-        /* We were provided with an outer frame; just ensure that it is
-         * based on the correct static frame (compare on bytecode address
-         * to cope with nqp::freshcoderef). */
+        /* We were provided with an outer frame and it will already have had
+         * its reference count incremented; just ensure that it is based on the
+         * correct static frame (compare on bytecode address to cope with
+         * nqp::freshcoderef). */
         if (outer->static_info->body.orig_bytecode == static_frame_body->outer->body.orig_bytecode)
             frame->outer = outer;
         else {
@@ -451,11 +525,12 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
         if (static_code->body.outer) {
             /* We're lacking an outer, but our static code object may have one.
             * This comes up in the case of cloned protoregexes, for example. */
-            frame->outer = static_code->body.outer;
+            frame->outer = MVM_frame_inc_ref_by_frame(tc, static_code->body.outer);
         }
         else if (static_frame_body->outer) {
             /* Auto-close, and cache it in the static frame. */
-            frame->outer = autoclose(tc, static_frame_body->outer);
+            frame->outer = MVM_frame_inc_ref_by_frame(tc,
+                autoclose(tc, static_frame_body->outer));
             static_code->body.outer = MVM_frame_inc_ref(tc, frame->outer);
         }
         else {
@@ -465,12 +540,10 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
     else {
         frame->outer = NULL;
     }
-    if (frame->outer)
-        MVM_frame_inc_ref(tc, frame->outer);
 
     /* Caller is current frame in the thread context. */
     if (tc->cur_frame)
-        frame->caller = MVM_frame_inc_ref(tc, tc->cur_frame);
+        frame->caller = MVM_frame_inc_ref_by_frame(tc, tc->cur_frame);
     else
         frame->caller = NULL;
     frame->keep_caller = 0;
@@ -576,7 +649,7 @@ MVMFrame * MVM_frame_create_for_deopt(MVMThreadContext *tc, MVMStaticFrame *stat
     frame->params.arg_flags         = NULL;
     frame->params.named_used        = NULL;
     if (code_ref->body.outer)
-        frame->outer = MVM_frame_inc_ref(tc, code_ref->body.outer);
+        frame->outer = MVM_frame_acquire_ref(tc, &(code_ref->body.outer));
     else
         frame->outer = NULL;
     return frame;
@@ -719,7 +792,7 @@ MVMuint64 MVM_frame_try_return(MVMThreadContext *tc) {
             default:
                 result = NULL;
         }
-        
+
         handler = MVM_frame_find_invokee(tc, hll->exit_handler, NULL);
         two_args_callsite = MVM_callsite_get_common(tc, MVM_CALLSITE_ID_TWO_OBJ);
         MVM_args_setup_thunk(tc, NULL, MVM_RETURN_VOID, two_args_callsite);
@@ -1100,12 +1173,18 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
     MVMuint32 ecost = 0;  /* frames traversed with empty cache */
     MVMuint32 xcost = 0;  /* frames traversed with wrong name */
     char *c_name;
+    MVMuint64 start_time;
+    MVMuint64 last_time;
 
     MVMFrame *initial_frame = cur_frame;
     if (!name)
         MVM_exception_throw_adhoc(tc, "Contextual name cannot be null");
-    if (dlog)
+    if (dlog) {
         c_name = MVM_string_utf8_encode_C_string(tc, name);
+        start_time = uv_hrtime();
+        last_time = tc->instance->dynvar_log_lasttime;
+    }
+
     MVM_string_flatten(tc, name);
     while (cur_frame != NULL) {
         MVMLexicalRegistry *lexical_names;
@@ -1136,9 +1215,10 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
                                 if (fcost+icost > 1)
                                   try_cache_dynlex(tc, initial_frame, cur_frame, name, result, *type, fcost, icost);
                                 if (dlog) {
-                                    fprintf(dlog, "I %s %d %d %d %d\n", c_name, fcost, icost, ecost, xcost);
+                                    fprintf(dlog, "I %s %d %d %d %d %llu %llu %llu\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
                                     fflush(dlog);
                                     MVM_free(c_name);
+                                    tc->instance->dynvar_log_lasttime = uv_hrtime();
                                 }
                                 return result;
                             }
@@ -1164,9 +1244,10 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
                                 if (fcost+icost > 1)
                                   try_cache_dynlex(tc, initial_frame, cur_frame, name, result, *type, fcost, icost);
                                 if (dlog) {
-                                    fprintf(dlog, "I %s %d %d %d %d\n", c_name, fcost, icost, ecost, xcost);
+                                    fprintf(dlog, "I %s %d %d %d %d %llu %llu %llu\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
                                     fflush(dlog);
                                     MVM_free(c_name);
+                                    tc->instance->dynvar_log_lasttime = uv_hrtime();
                                 }
                                 return result;
                             }
@@ -1186,9 +1267,10 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
                 if (fcost+icost > 5)
                     try_cache_dynlex(tc, initial_frame, cur_frame, name, result, *type, fcost, icost);
                 if (dlog) {
-                    fprintf(dlog, "C %s %d %d %d %d\n", c_name, fcost, icost, ecost, xcost);
+                    fprintf(dlog, "C %s %d %d %d %d %llu %llu %llu\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
                     fflush(dlog);
                     MVM_free(c_name);
+                    tc->instance->dynvar_log_lasttime = uv_hrtime();
                 }
                 return result;
             }
@@ -1208,9 +1290,10 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
                 if (vivify && *type == MVM_reg_obj && !result->o)
                     MVM_frame_vivify_lexical(tc, cur_frame, entry->value);
                 if (dlog) {
-                    fprintf(dlog, "F %s %d %d %d %d\n", c_name, fcost, icost, ecost, xcost);
+                    fprintf(dlog, "F %s %d %d %d %d %llu %llu %llu\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
                     fflush(dlog);
                     MVM_free(c_name);
+                    tc->instance->dynvar_log_lasttime = uv_hrtime();
                 }
                 if (fcost+icost > 1)
                     try_cache_dynlex(tc, initial_frame, cur_frame, name, result, *type, fcost, icost);
@@ -1221,9 +1304,10 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
         cur_frame = cur_frame->caller;
     }
     if (dlog) {
-        fprintf(dlog, "N %s %d %d %d %d\n", c_name, fcost, icost, ecost, xcost);
+        fprintf(dlog, "N %s %d %d %d %d %llu %llu %llu\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
         fflush(dlog);
         MVM_free(c_name);
+        tc->instance->dynvar_log_lasttime = uv_hrtime();
     }
     return NULL;
 }
@@ -1404,8 +1488,9 @@ static MVMObject * find_invokee_internal(MVMThreadContext *tc, MVMObject *code,
             }
             else {
                 MVMCallsite *new   = MVM_malloc(sizeof(MVMCallsite));
-                MVMint32     fsize = orig->num_pos + (orig->arg_count - orig->num_pos) / 2;
-                new->arg_flags     = MVM_malloc((fsize + 1) * sizeof(MVMCallsiteEntry));
+                MVMint32     fsize = orig->flag_count;
+                new->flag_count    = fsize + 1;
+                new->arg_flags     = MVM_malloc(new->flag_count * sizeof(MVMCallsiteEntry));
                 new->arg_flags[0]  = MVM_CALLSITE_ARG_OBJ;
                 memcpy(new->arg_flags + 1, orig->arg_flags, fsize);
                 new->arg_count      = orig->arg_count + 1;
@@ -1450,8 +1535,7 @@ MVMObject * MVM_frame_find_invokee_multi_ok(MVMThreadContext *tc, MVMObject *cod
     if (STABLE(code)->invoke == MVM_6model_invoke_default) {
         MVMInvocationSpec *is = STABLE(code)->invocation_spec;
         if (!is) {
-            MVM_exception_throw_adhoc(tc, "Cannot invoke this object (REPR: %s, cs = %d)",
-                REPR(code)->name, STABLE(code)->container_spec ? 1 : 0);
+            MVM_exception_throw_adhoc(tc, "Cannot invoke this object");
         }
         if (!MVM_is_null(tc, is->md_class_handle)) {
             /* We might be able to dig straight into the multi cache and not
