@@ -77,6 +77,8 @@ MVMint16 MVM_nativecall_get_arg_type(MVMThreadContext *tc, MVMObject *info, MVMi
         result = MVM_NATIVECALL_ARG_UTF16STR | get_str_free_flag(tc, info);
     else if (strcmp(ctypename, "cstruct") == 0)
         result = MVM_NATIVECALL_ARG_CSTRUCT;
+    else if (strcmp(ctypename, "cppstruct") == 0)
+        result = MVM_NATIVECALL_ARG_CPPSTRUCT;
     else if (strcmp(ctypename, "cpointer") == 0)
         result = MVM_NATIVECALL_ARG_CPOINTER;
     else if (strcmp(ctypename, "carray") == 0)
@@ -169,6 +171,21 @@ MVMObject * MVM_nativecall_make_cunion(MVMThreadContext *tc, MVMObject *type, vo
     return result;
 }
 
+MVMObject * MVM_nativecall_make_cppstruct(MVMThreadContext *tc, MVMObject *type, void *cppstruct) {
+    MVMObject *result = type;
+    if (cppstruct && type) {
+        MVMCPPStructREPRData *repr_data = (MVMCPPStructREPRData *)STABLE(type)->REPR_data;
+        if (REPR(type)->ID != MVM_REPR_ID_MVMCPPStruct)
+            MVM_exception_throw_adhoc(tc,
+                "Native call expected return type with CPPStruct representation, but got a %s", REPR(type)->name);
+        result = REPR(type)->allocate(tc, STABLE(type));
+        ((MVMCPPStruct *)result)->body.cppstruct = cppstruct;
+        if (repr_data->num_child_objs)
+            ((MVMCPPStruct *)result)->body.child_objs = MVM_calloc(repr_data->num_child_objs, sizeof(MVMObject *));
+    }
+    return result;
+}
+
 /* Constructs a boxed result using a CPointer REPR type. */
 MVMObject * MVM_nativecall_make_cpointer(MVMThreadContext *tc, MVMObject *type, void *ptr) {
     MVMObject *result = type;
@@ -287,6 +304,16 @@ void * MVM_nativecall_unmarshal_cstruct(MVMThreadContext *tc, MVMObject *value)
             "Native call expected return type with CStruct representation, but got a %s", REPR(value)->name);
 }
 
+void * MVM_nativecall_unmarshal_cppstruct(MVMThreadContext *tc, MVMObject *value) {
+    if (!IS_CONCRETE(value))
+        return NULL;
+    else if (REPR(value)->ID == MVM_REPR_ID_MVMCPPStruct)
+        return ((MVMCPPStruct *)value)->body.cppstruct;
+    else
+        MVM_exception_throw_adhoc(tc,
+            "Native call expected return type with CPPStruct representation, but got a %s", REPR(value)->name);
+}
+
 void * MVM_nativecall_unmarshal_cpointer(MVMThreadContext *tc, MVMObject *value) {
     if (!IS_CONCRETE(value))
         return NULL;
@@ -346,8 +373,8 @@ static const char *dlerror(void)
 /* Builds up a native call site out of the supplied arguments. */
 void MVM_nativecall_build(MVMThreadContext *tc, MVMObject *site, MVMString *lib,
         MVMString *sym, MVMString *conv, MVMObject *arg_info, MVMObject *ret_info) {
-    char *lib_name = MVM_string_utf8_encode_C_string(tc, lib);
-    char *sym_name = MVM_string_utf8_encode_C_string(tc, sym);
+    char *lib_name = MVM_string_utf8_c8_encode_C_string(tc, lib);
+    char *sym_name = MVM_string_utf8_c8_encode_C_string(tc, sym);
     MVMint16 i;
 
     /* Initialize the object; grab native call part of its body. */
@@ -414,20 +441,36 @@ static MVMObject * nativecall_cast(MVMThreadContext *tc, MVMObject *target_spec,
                     const MVMStorageSpec *ss = REPR(target_spec)->get_storage_spec(tc, STABLE(target_spec));
                     if(ss->can_box & MVM_STORAGE_SPEC_CAN_BOX_INT) {
                         MVMint64 value = 0;
-                        switch(ss->bits) {
-                            case 8:
-                                value = *(MVMint8 *)cpointer_body;
-                                break;
-                            case 16:
-                                value = *(MVMint16 *)cpointer_body;
-                                break;
-                            case 32:
-                                value = *(MVMint32 *)cpointer_body;
-                                break;
-                            case 64:
-                            default:
-                                value = *(MVMint64 *)cpointer_body;
-                        }
+                        if (ss->is_unsigned)
+                            switch(ss->bits) {
+                                case 8:
+                                    value = *(MVMuint8 *)cpointer_body;
+                                    break;
+                                case 16:
+                                    value = *(MVMuint16 *)cpointer_body;
+                                    break;
+                                case 32:
+                                    value = *(MVMuint32 *)cpointer_body;
+                                    break;
+                                case 64:
+                                default:
+                                    value = *(MVMuint64 *)cpointer_body;
+                            }
+                        else
+                            switch(ss->bits) {
+                                case 8:
+                                    value = *(MVMint8 *)cpointer_body;
+                                    break;
+                                case 16:
+                                    value = *(MVMint16 *)cpointer_body;
+                                    break;
+                                case 32:
+                                    value = *(MVMint32 *)cpointer_body;
+                                    break;
+                                case 64:
+                                default:
+                                    value = *(MVMint64 *)cpointer_body;
+                            }
 
                         result = MVM_nativecall_make_int(tc, target_type, value);
                     }
@@ -454,20 +497,36 @@ static MVMObject * nativecall_cast(MVMThreadContext *tc, MVMObject *target_spec,
                 case MVM_REPR_ID_P6int: {
                     const MVMStorageSpec *ss = REPR(target_spec)->get_storage_spec(tc, STABLE(target_spec));
                     MVMint64 value;
-                    switch(ss->bits) {
-                        case 8:
-                            value = *(MVMint8 *)cpointer_body;
-                            break;
-                        case 16:
-                            value = *(MVMint16 *)cpointer_body;
-                            break;
-                        case 32:
-                            value = *(MVMint32 *)cpointer_body;
-                            break;
-                        case 64:
-                        default:
-                            value = *(MVMint64 *)cpointer_body;
-                    }
+                    if (ss->is_unsigned)
+                        switch(ss->bits) {
+                            case 8:
+                                value = *(MVMuint8 *)cpointer_body;
+                                break;
+                            case 16:
+                                value = *(MVMuint16 *)cpointer_body;
+                                break;
+                            case 32:
+                                value = *(MVMuint32 *)cpointer_body;
+                                break;
+                            case 64:
+                            default:
+                                value = *(MVMuint64 *)cpointer_body;
+                        }
+                    else
+                        switch(ss->bits) {
+                            case 8:
+                                value = *(MVMint8 *)cpointer_body;
+                                break;
+                            case 16:
+                                value = *(MVMint16 *)cpointer_body;
+                                break;
+                            case 32:
+                                value = *(MVMint32 *)cpointer_body;
+                                break;
+                            case 64:
+                            default:
+                                value = *(MVMint64 *)cpointer_body;
+                        }
                     result = MVM_nativecall_make_int(tc, target_type, value);
                     break;
                 }
@@ -515,8 +574,8 @@ static MVMObject * nativecall_cast(MVMThreadContext *tc, MVMObject *target_spec,
 }
 
 MVMObject * MVM_nativecall_global(MVMThreadContext *tc, MVMString *lib, MVMString *sym, MVMObject *target_spec, MVMObject *target_type) {
-    char *lib_name = MVM_string_utf8_encode_C_string(tc, lib);
-    char *sym_name = MVM_string_utf8_encode_C_string(tc, sym);
+    char *lib_name = MVM_string_utf8_c8_encode_C_string(tc, lib);
+    char *sym_name = MVM_string_utf8_c8_encode_C_string(tc, sym);
     DLLib *lib_handle;
     void *entry_point;
     MVMObject *ret = NULL;
@@ -559,6 +618,8 @@ MVMObject * MVM_nativecall_cast(MVMThreadContext *tc, MVMObject *target_spec, MV
 
     if (REPR(source)->ID == MVM_REPR_ID_MVMCStruct)
         data_body = MVM_nativecall_unmarshal_cstruct(tc, source);
+    else if (REPR(source)->ID == MVM_REPR_ID_MVMCPPStruct)
+        data_body = MVM_nativecall_unmarshal_cppstruct(tc, source);
     else if (REPR(source)->ID == MVM_REPR_ID_MVMCUnion)
         data_body = MVM_nativecall_unmarshal_cunion(tc, source);
     else if (REPR(source)->ID == MVM_REPR_ID_MVMCPointer)
@@ -576,6 +637,8 @@ MVMObject * MVM_nativecall_cast(MVMThreadContext *tc, MVMObject *target_spec, MV
 MVMint64 MVM_nativecall_sizeof(MVMThreadContext *tc, MVMObject *obj) {
     if (REPR(obj)->ID == MVM_REPR_ID_MVMCStruct)
         return ((MVMCStructREPRData *)STABLE(obj)->REPR_data)->struct_size;
+    else if (REPR(obj)->ID == MVM_REPR_ID_MVMCPPStruct)
+        return ((MVMCPPStructREPRData *)STABLE(obj)->REPR_data)->struct_size;
     else if (REPR(obj)->ID == MVM_REPR_ID_MVMCUnion)
         return ((MVMCUnionREPRData *)STABLE(obj)->REPR_data)->struct_size;
     else if (REPR(obj)->ID == MVM_REPR_ID_P6int)
@@ -695,6 +758,52 @@ void MVM_nativecall_refresh(MVMThreadContext *tc, MVMObject *cthingy) {
                 objptr = NULL;
             }
 
+            if (objptr != cptr)
+                body->child_objs[slot] = NULL;
+            else
+                MVM_nativecall_refresh(tc, body->child_objs[slot]);
+        }
+    }
+    else if (REPR(cthingy)->ID == MVM_REPR_ID_MVMCPPStruct) {
+        MVMCPPStructBody     *body      = (MVMCPPStructBody *)OBJECT_BODY(cthingy);
+        MVMCPPStructREPRData *repr_data = (MVMCPPStructREPRData *)STABLE(cthingy)->REPR_data;
+        char                 *storage   = (char *) body->cppstruct;
+        MVMint64              i;
+
+        for (i = 0; i < repr_data->num_attributes; i++) {
+            MVMint32 kind = repr_data->attribute_locations[i] & MVM_CPPSTRUCT_ATTR_MASK;
+            MVMint32 slot = repr_data->attribute_locations[i] >> MVM_CPPSTRUCT_ATTR_SHIFT;
+            void *cptr;   /* The pointer in the C storage. */
+            void *objptr; /* The pointer in the object representing the C object. */
+
+            if (kind == MVM_CPPSTRUCT_ATTR_IN_STRUCT || !body->child_objs[slot])
+                continue;
+
+            cptr = *((void **)(storage + repr_data->struct_offsets[i]));
+            if (IS_CONCRETE(body->child_objs[slot])) {
+                switch (kind) {
+                    case MVM_CPPSTRUCT_ATTR_CARRAY:
+                        objptr = ((MVMCArrayBody *)OBJECT_BODY(body->child_objs[slot]))->storage;
+                        break;
+                    case MVM_CPPSTRUCT_ATTR_CPTR:
+                        objptr = ((MVMCPointerBody *)OBJECT_BODY(body->child_objs[slot]))->ptr;
+                        break;
+                    case MVM_CPPSTRUCT_ATTR_CSTRUCT:
+                        objptr = (MVMCStructBody *)OBJECT_BODY(body->child_objs[slot]);
+                        break;
+                    case MVM_CPPSTRUCT_ATTR_STRING:
+                        objptr = NULL;
+                        break;
+                    default:
+                        MVM_exception_throw_adhoc(tc,
+                            "Fatal error: bad kind (%d) in CPPStruct write barrier",
+                            kind);
+                }
+            }
+            else {
+                objptr = NULL;
+            }
+
             if (objptr != cptr)
                 body->child_objs[slot] = NULL;
             else
