@@ -271,7 +271,7 @@ void MVM_gc_mark_thread_unblocked(MVMThreadContext *tc) {
 
 static MVMint32 is_full_collection(MVMThreadContext *tc) {
     MVMuint64 threshold = MVM_GC_GEN2_THRESHOLD_BASE +
-        (tc->instance->num_user_threads * MVM_GC_GEN2_THRESHOLD_THREAD);
+        ((MVMuint64)tc->instance->num_user_threads * MVM_GC_GEN2_THRESHOLD_THREAD);
     return MVM_load(&tc->instance->gc_promoted_bytes_since_last_full) > threshold;
 }
 
@@ -318,7 +318,7 @@ static void run_gc(MVMThreadContext *tc, MVMuint8 what_to_do) {
             GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
                 "Thread %d run %d : freeing gen2 of thread %d\n",
                 other->thread_id);
-            MVM_gc_collect_free_gen2_unmarked(other);
+            MVM_gc_collect_free_gen2_unmarked(other, 0);
         }
     }
 }
@@ -410,6 +410,13 @@ void MVM_gc_enter_from_allocator(MVMThreadContext *tc) {
         if (is_full)
             MVM_store(&tc->instance->gc_promoted_bytes_since_last_full, 0);
 
+        /* This is a safe point for us to free any STables that have been marked
+         * for deletion in the previous collection (since we let finalization -
+         * which appends to this list - happen after we set threads on their
+         * way again, it's not safe to do it in the previous collection). */
+        GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE, "Thread %d run %d : Freeing STables if needed\n");
+        MVM_gc_collect_free_stables(tc);
+
         /* Signal to the rest to start */
         GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE, "Thread %d run %d : coordinator signalling start\n");
         if (MVM_decr(&tc->instance->gc_start) != 1)
@@ -419,12 +426,6 @@ void MVM_gc_enter_from_allocator(MVMThreadContext *tc) {
         GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE, "Thread %d run %d : coordinator entering run_gc\n");
         run_gc(tc, MVMGCWhatToDo_All);
 
-        /* Free any STables that have been marked for deletion. It's okay for
-         * us to muck around in another thread's fromspace while it's mutating
-         * tospace, really. */
-        GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE, "Thread %d run %d : Freeing STables if needed\n");
-        MVM_gc_collect_free_stables(tc);
-
         /* If profiling, record that GC is over. */
         if (tc->instance->profiling)
             MVM_profiler_log_gc_end(tc);
@@ -498,6 +499,6 @@ void MVM_gc_global_destruction(MVMThreadContext *tc) {
     /* Run the objects' finalizers */
     MVM_gc_collect_free_nursery_uncopied(tc, tc->nursery_alloc);
     MVM_gc_root_gen2_cleanup(tc);
-    MVM_gc_collect_free_gen2_unmarked(tc);
+    MVM_gc_collect_free_gen2_unmarked(tc, 1);
     MVM_gc_collect_free_stables(tc);
 }
