@@ -6,13 +6,13 @@
  * in http://www.cs.rice.edu/~keith/EMBED/dom.pdf. The SSA algorithm itself is
  * from http://www.cs.utexas.edu/~pingali/CS380C/2010/papers/ssaCytron.pdf. */
 
-#define GET_I8(pc, idx)     *((MVMint8 *)(pc + idx))
-#define GET_UI8(pc, idx)    *((MVMuint8 *)(pc + idx))
-#define GET_I16(pc, idx)    *((MVMint16 *)(pc + idx))
-#define GET_UI16(pc, idx)   *((MVMuint16 *)(pc + idx))
-#define GET_I32(pc, idx)    *((MVMint32 *)(pc + idx))
-#define GET_UI32(pc, idx)   *((MVMuint32 *)(pc + idx))
-#define GET_N32(pc, idx)    *((MVMnum32 *)(pc + idx))
+#define GET_I8(pc, idx)     *((MVMint8 *)((pc) + (idx)))
+#define GET_UI8(pc, idx)    *((MVMuint8 *)((pc) + (idx)))
+#define GET_I16(pc, idx)    *((MVMint16 *)((pc) + (idx)))
+#define GET_UI16(pc, idx)   *((MVMuint16 *)((pc) + (idx)))
+#define GET_I32(pc, idx)    *((MVMint32 *)((pc) + (idx)))
+#define GET_UI32(pc, idx)   *((MVMuint32 *)((pc) + (idx)))
+#define GET_N32(pc, idx)    *((MVMnum32 *)((pc) + (idx)))
 
 /* Allocate a piece of memory from the spesh graph's buffer. Deallocated when
  * the spesh graph is. */
@@ -257,8 +257,10 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
                         "Spesh: unknown operand type %d in graph building (op %s)",
                         (int)type, ins_node->info->name);
                 }
+                break;
+            default:
+                break;
             }
-            break;
             }
         }
 
@@ -298,6 +300,8 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
             byte_to_ins_flags[pc - g->bytecode] |= MVM_CFG_BB_END;
             next_bbs = 1;
             break;
+        default:
+            break;
         }
 
         /* Final instruction is basic block end. */
@@ -618,7 +622,9 @@ static MVMSpeshBB ** reverse_postorder(MVMThreadContext *tc, MVMSpeshGraph *g) {
     dfs(rpo, &ins, seen, g->entry);
     MVM_free(seen);
     if (ins != -1) {
-        printf("%s", MVM_spesh_dump(tc, g));
+        char *dump_msg = MVM_spesh_dump(tc, g);
+        printf("%s", dump_msg);
+        MVM_free(dump_msg);
         MVM_spesh_graph_destroy(tc, g);
         MVM_oops(tc, "Spesh: reverse postorder calculation failed");
     }
@@ -630,7 +636,9 @@ static void iter_check(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB **rpo,
     if (iters > 100000) {
 #ifdef NDEBUG
         MVMint32 k;
-        printf("%s", MVM_spesh_dump(tc, g));
+        char *dump_msg = MVM_spesh_dump(tc, g);
+        printf("%s", dump_msg);
+        MVM_free(dump_msg);
         printf("RPO: ");
         for (k = 0; k < g->num_bbs; k++)
             printf("%d, ", rpo[k]->idx);
